# functions that do configuration, session management + generic functions

import os
import sys

import yaml
from shutil import copy2 as cp
import json

__all__ = ['encode', 'config', 'session']

from rubecula.logging import console


def encode(msg):
    if not isinstance(msg, bytes):
        if not isinstance(msg, str):
            msg = json.dumps(msg)
        msg = msg.encode()
    return msg


def get_config():
    if not os.path.exists('config.yaml'):
        if os.path.exists("config.yaml.example"):
            cp('config.yaml.example', 'config.yaml')
        else:
            console.error("No config.yaml or config.yaml.example could be found. "
                          "Please create one in the project's root folder!")
            exit(1)

    with open('config.yaml') as stream:
        yaml_config = yaml.load(stream, Loader=yaml.FullLoader)

    console.loglevel = yaml_config.get('log')

    return yaml_config


def write_config(c_dict):
    with open('config.yaml', 'w') as f:
        yaml.dump(c_dict, f, default_flow_style=False)


config = get_config()

session = {
}  # session storage



from functools import partial

import andi


class Inject:
    pass


def _build(plan, instances_stock=None):
    instances_stock = instances_stock or {}
    instances = {}
    for fn_or_cls, kwargs_spec in plan:
        if fn_or_cls in instances_stock:
            instances[fn_or_cls] = instances_stock[fn_or_cls]
        else:
            instances[fn_or_cls] = fn_or_cls(**kwargs_spec.kwargs(instances))
    return instances


def inject_everything(_):
    # return True
    return issubclass(_, Inject)


def get_instance(cls, ext=None):
    plan = andi.plan(cls, is_injectable=inject_everything, externally_provided=ext)
    instances = _build(plan)
    return instances[cls]


def get_method(func, dependencies=None):
    externally_provided = {c.__class__ for c in dependencies}
    instances = {c.__class__: c for c in dependencies}

    plan = andi.plan(func, is_injectable=inject_everything, externally_provided=externally_provided)
    instances = _build(plan.dependencies, instances)
    return partial(func, **plan.final_kwargs(instances))

# class Demo(Inject):
#     def __init__(self):
#        pass
#
#
# def vroem(speed: int, d: Demo):
#     print(type(speed))
#     print("I'm in my moms", d)
#     print(f"Driving {speed} mph")
#
# _vroem = get_method(vroem)
# _vroem("100")



from rich.console import Console
from rich.theme import Theme

__all__ = ["console"]

custom_theme = Theme({
    "error": "bold red",
    "warn": "bold yellow",
    "info": "dim cyan",
    "success": "bold green",
    "debug": "dim magenta",
})


class RubeculaConsole(Console):
    """
    @DynamicAttrs
    """

    def __init__(self, *a, **kw):
        super().__init__(*a, **kw, theme=custom_theme)
        self.loglevel = 3
        # self.loglevel = 100

        for i, level in enumerate([
            'error', 'warn', 'info', 'success', 'debug'
        ]):
            self.__setattr__(level, self.__log_with_level(level, i))

    def __log_with_level(self, level, i):
        def func(*a, **kw):
            if i > self.loglevel:
                return

            self.print(*a, **kw, style=level)

        return func


console = RubeculaConsole()

# console.error("---")
# console.warn("Welcome")
# console.info("to")
# console.success("Rubecula")
# console.debug("---")



import sys

from autobahn.twisted.websocket import WebSocketServerFactory  # WebSocketServerProtocol
from twisted.internet import reactor, ssl
from twisted.python import log
from twisted.web.server import Site
from autobahn.twisted.resource import WebSocketResource

from .config import config
from .logging import console
from .ws import Server
from .web import root
from ._autoreload import autoreload


def _create_listener(port, site_root, ssl_factory):
    if ssl_factory:
        return reactor.listenSSL(int(port), Site(site_root), ssl_factory)
    else:
        return reactor.listenTCP(int(port), Site(site_root))


def _serve(arguments):
    if not arguments:
        port = '8080'
    else:
        port = arguments[0]
    log.startLogging(sys.stdout)
    if config.get('ssl'):
        factory = WebSocketServerFactory(u"wss://127.0.0.1:{}".format(port))
        ssl_factory = ssl.DefaultOpenSSLContextFactory('keys/privkey.pem',
                                                       'keys/cert.pem')
    else:
        factory = WebSocketServerFactory("ws://127.0.0.1:{}".format(port))
        ssl_factory = None

    factory.protocol = Server

    # run our WebSocket server under "/ws" (note that Twisted uses bytes for URIs)
    root.putChild(b"ws", WebSocketResource(factory))
    # start listening
    _create_listener(port, root, ssl_factory)
    reactor.run()


def serve(args, dev=None):
    if dev:
        import time
        try:
            autoreload(_serve, args)
        except Exception as e:
            console.error(e)
            time.sleep(3)
            serve(args, dev)
    else:
        _serve(args)



# functions for exposing websocket endpoints

from autobahn.twisted.websocket import WebSocketServerProtocol

from . import injection
from .config import encode
import json
from dataclasses import dataclass

__all__ = ['expose_ws', 'peers', 'Server', "Client", "js", 'functions']

from .injection import Inject

peers, functions = {}, {}


def expose_ws(command=None):
    def wrapper(f):

        if command and not callable(command):
            cmd = command
        else:
            cmd = f.__name__

        def inner_wrapper(*a, **kw):
            f(*a, **kw)

        functions[cmd] = f
        return inner_wrapper

    if callable(command):
        return wrapper(command)
    else:
        return wrapper


class Server(WebSocketServerProtocol, Inject):

    def __init__(self):
        self.client = Client(self)
        super().__init__()

    @property
    def peers(self):
        return peers.items()

    @staticmethod
    def default_func(*_):
        raise NotImplementedError('please query another method ({})'.format(list(functions)))

    @staticmethod
    def payload_to_dict(pl):
        return json.loads(pl.decode())

    def send_client(self, message):
        self.sendMessage(encode(message))

    def _handle(self, payload):
        payload = self.payload_to_dict(payload)
        function = functions.get(payload.get('function', None), self.default_func)

        function = injection.get_method(function, dependencies=[self, self.client])

        result = function(payload.get('data', {}))

        if payload.get('return'):
            # add 'return' token to response:
            if isinstance(result, dict):
                result['return'] = payload['return']
            else:
                result = {
                    'return': payload['return'],
                    'data': result,
                }

        return encode(result)

    def onConnect(self, request):
        # add peer to peer pool
        peers[request.peer] = self

    def onMessage(self, payload, is_binary):
        try:
            self.send_client(self._handle(payload))
        except Exception as e:
            self.send_client('An error occured: {}'.format(str(repr(e))))

    def onClose(self, was_clean, code, reason):
        if peers.get(self.peer):
            # cleanup
            del peers[self.peer]

    @property
    def js(self):
        return JavaScript(server=self)


@dataclass
class JavascriptCall:
    func: str
    server: Server = None

    def __call__(self, data, **kwargs):
        data = {
            'function': self.func,
            'data': data,
            **kwargs
        }

        if self.server:
            return self.server.send_client(data)
        else:
            return data


@dataclass
class JavaScript:
    server: Server = None

    def __getattr__(self, key):
        return JavascriptCall(key, self.server)


@dataclass
class Client(JavaScript, Inject):
    # dummy for ease of use
    server: Server = None


js = JavaScript()



# functions for exposing web endpoints
from twisted.web.server import Request
from twisted.web.resource import Resource
from twisted.web.static import File
from jinja2 import Environment, FileSystemLoader
from .config import encode
import json

__all__ = ['expose_static', 'expose_web', 'root', "Request", "Args"]

from .injection import Inject, get_method


def HTTP_CODE(code, message='', request=None):
    if request:
        request.setResponseCode(code)
    return encode(f'{message}')


# we server static files under "/view"

root = File("view")


class StaticFile(File):
    def __init__(self, trigger_func, *a, **kw):
        self.trigger_func = trigger_func
        super().__init__(*a, **kw)

    def render(self, request, *a, **kw):
        args = Args(request.args)
        f = get_method(self.trigger_func, [request, args])

        res = super().render(request)
        f(*a, **kw)
        return res


class Args(dict, Inject):
    def __init__(self, *a, **kw):
        super().__init__(*a, **kw)
        self._convert()

    @classmethod
    def __convert(cls, data):
        if isinstance(data, bytes):
            return data.decode('ascii')
        if isinstance(data, dict):
            return dict(map(cls.__convert, data.items()))
        if isinstance(data, (tuple, list)):
            return type(data)(map(cls.__convert, data))

        return data

    def _convert(self):
        new = self.__convert(self)
        self.clear()
        self.update(new)


class Template(Resource):
    def __init__(self, func, template):
        super().__init__()

        if '.' not in template:
            template = template + '.html'

        self.template = template
        self.function = func

    def render(self, request):
        args = Args(request.args)
        func = get_method(self.function, [request, args])

        ctx = func()
        ctx["__globals"] = json.dumps(ctx)
        env = Environment(loader=FileSystemLoader('view'))

        template = env.get_template(self.template)

        output_from_parsed_template = template.render(**ctx)
        return encode(output_from_parsed_template)


def expose_static(alias=None, template=True):
    def decorator(func):
        def wrapper():
            name = func.__name__

            filename = name if template in (True, False, None) else template

            if '.' not in filename:
                filename += '.html'

            if alias and not callable(alias):
                where = alias
            else:
                where = name

            if template:
                contents = Template(func, filename)
            else:
                contents = StaticFile(func, f'view/{filename}')

            root.putChild(where.encode(), contents)

        return wrapper()

    if callable(alias):
        return decorator(alias)  # return 'wrapper'
    else:
        return decorator  # ... or 'decorator'


def _safe_execute(func, request):
    try:
        args = Args(request.args)
        func = get_method(func, dependencies=[request, args])

        return encode(func())
    except Exception as e:
        return HTTP_CODE(400, f'Something went wrong: {e}', request)


def _add_page(root, name, function_post=None, function_get=None):
    class Page(Resource):
        def render_POST(self, request):
            if not function_post:
                return HTTP_CODE(405, "Method 'POST' not allowed.", request)
            return _safe_execute(function_post, request)

        def render_GET(self, request):
            if not function_get:
                return HTTP_CODE(405, "Method 'GET' not allowed.", request)
            return _safe_execute(function_get, request)

    p = Page()

    # allow trailing slash:
    p.putChild(b"", p)

    root.putChild(name.encode(), p)


def expose_web(method=None, alias=None):
    def wrapper(f):
        def inner_wrapper(*a, **kw):
            f(*a, **kw)

        request_method = method.lower() if method and isinstance(method, str) else 'get'

        kwargs = {}

        if request_method == 'post':
            kwargs['function_post'] = f
        elif request_method == 'get':
            kwargs['function_get'] = f
        else:
            raise NotImplementedError(f'Method {request_method} not implemented.')

        _add_page(root, f.__name__, **kwargs)

        if f.__name__ == "index":
            # index is special
            _add_page(root, "", **kwargs)

        if alias is not None:
            if not isinstance(alias, (tuple, list)):
                aliases = [alias]
            else:
                aliases = alias

            for a in aliases:
                _add_page(root, a, **kwargs)

        return inner_wrapper

    if callable(method):
        return wrapper(method)
    else:
        return wrapper



from functools import partial

from .ws import expose_ws
from .web import expose_web
from .web import expose_static

__all__ = ["websocket", "web", "static", "template"]

websocket = expose_ws
web = expose_web
template = expose_static

static = partial(template, template=False)



from setuptools import setup

setup(
    name='rubecula',
    version='0.1.2',
    packages=['rubecula'],
    install_requires=[
        'twisted',
        'autobahn',
        'requests',
        'pyopenssl',
        'service_identity',
        'pyyaml',
        'jinja2',
        'andi',
        'rich',
    ],
)



#
# usage:
#
# import autoreload
#
# def myMain():
#    site = server.Site(root)
#    reactor.listenTCP(8080, site)
#    reactor.run()
#
# autoreload.main(myMain)
#
# Autoreloading launcher.
# Borrowed from Peter Hunt and the CherryPy project (http://www.cherrypy.org).
# Some taken from Ian Bicking's Paste (http://pythonpaste.org/).
# Adjustments made by Michael Elsdoerfer (michael@elsdoerfer.com).
#
# Portions copyright (c) 2004, CherryPy Team (team@cherrypy.org)
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
#     * Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright notice,
#       this list of conditions and the following disclaimer in the documentation
#       and/or other materials provided with the distribution.
#     * Neither the name of the CherryPy Team nor the names of its contributors
#       may be used to endorse or promote products derived from this software
#       without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# https://gist.github.com/motleytech/8f255193f613c6623c19d3f93c01cbed

import os, sys, time

import _thread as thread

# This import does nothing, but it's necessary to avoid some race conditions
# in the threading module. See http://code.djangoproject.com/ticket/2330 .
try:
    import threading
except ImportError:
    pass

RUN_RELOADER = True

_mtimes = {}
_win = (sys.platform == "win32")


def code_changed():
    global _mtimes, _win

    code_files = list(filter(lambda v: v, map(lambda m: getattr(m, "__file__", None), sys.modules.values())))
    code_files.extend(['config.yaml', 'config.example.yaml'])

    for filename in code_files:
        if filename.endswith(".pyc") or filename.endswith(".pyo"):
            filename = filename[:-1]
        if not os.path.exists(filename):
            continue  # File might be in an egg, so it can't be reloaded.
        stat = os.stat(filename)
        mtime = stat.st_mtime
        if _win:
            mtime -= stat.st_ctime
        if filename not in _mtimes:
            _mtimes[filename] = mtime
            continue
        if mtime != _mtimes[filename]:
            _mtimes = {}
            return True
    return False


def reloader_thread(softexit=False):
    """If ``soft_exit`` is True, we use sys.exit(); otherwise ``os_exit``
    will be used to end the process.
    """
    while RUN_RELOADER:
        if code_changed():
            # force reload
            if softexit:
                sys.exit(3)
            else:
                os._exit(3)
        time.sleep(1)


def restart_with_reloader():
    while True:
        args = [sys.executable] + sys.argv
        if sys.platform == "win32":
            args = ['"%s"' % arg for arg in args]
        new_environ = os.environ.copy()
        new_environ["RUN_MAIN"] = 'true'
        exit_code = os.spawnve(os.P_WAIT, sys.executable, args, new_environ)
        if exit_code != 3:
            return exit_code


def python_reloader(main_func, args, kwargs, check_in_thread=True):
    """
    If ``check_in_thread`` is False, ``main_func`` will be run in a separate
    thread, and the code checker in the main thread. This was the original
    behavior of this module: I (Michael Elsdoerfer) changed the default
    to be the reverse: Code checker in thread, main func in main thread.
    This was necessary to make the thing work with Twisted
    (http://twistedmatrix.com/trac/ticket/4072).
    """
    if os.environ.get("RUN_MAIN") == "true":
        if check_in_thread:
            thread.start_new_thread(reloader_thread, (), {'softexit': False})
        else:
            thread.start_new_thread(main_func, args, kwargs)

        try:
            if not check_in_thread:
                reloader_thread(softexit=True)
            else:
                main_func(*args, **kwargs)
        except KeyboardInterrupt:
            pass
    else:
        try:
            sys.exit(restart_with_reloader())
        except KeyboardInterrupt:
            pass


def jython_reloader(main_func, args, kwargs):
    from _systemrestart import SystemRestart
    thread.start_new_thread(main_func, args)
    while True:
        if code_changed():
            raise SystemRestart
        time.sleep(1)


def autoreload(main_func, args=None, kwargs=None, **more_options):
    if args is None:
        args = ()
    if kwargs is None:
        kwargs = {}
    if sys.platform.startswith('java'):
        reloader = jython_reloader
    else:
        reloader = python_reloader
    reloader(main_func, args, kwargs, **more_options)


main = autoreload
